/*
**   PARFIMA.PRG
**  Purpose: Estimation an Forecasting of a PARFIMA (0,Ds,0) model
**  Method : Approximate ML, (Beran (JRSS,1995)),allowing for outliers in a
**           simple way)
**  Author: Marius Ooms
**  Files:  nroptmum.ext, nroptmum.dec, nroptmum.src (by Honor and Kyriazidou)
**          parfima.ext, parfima.dec, parfima.src
**
**  For updates, errors and comments:
**   http://www.few.eur.nl/few/ei/links/ooms
**   e-mail: ooms@few.eur.nl
**  Datafile:  rpiinfne.dat
**  Date:   13/9/96
**  GAUSS for DOS version: 3.2.13, GAUSS for Unix 3.2.7
**
**  Reference: Franses, P.H., and Ooms, M. (1996), ''A Periodic Long Memory
**             Model for Quarterly UK Inflation'', International
**             Journal of Forecasting, 13, issue 1, pages 119-128
*/
library pgraph;
#include nroptmum.ext;@ get this from nroptmum library http://www.eur.nl/few/ei/links/ooms/programs/nroptmum.zip @
#include nroptmum.dec;@  get this from nroptmum library http://www.eur.nl/few/ei/links/ooms/programs/nroptmum.zip @
#include nroptmum.src;@ get these from nroptmum library http://www.eur.nl/few/ei/links/ooms/programs/nroptmum.zip @
                      @ you can make your own library file to speed things up! @
#include parfima.dec;
#include parfima.ext;
#include parfima.src;@ you can make your own library file to speed things up! @
/*
** Data
*/
datafile="rpiinfne.dat";
outfile ="parfiest.lis";
load x[]=^datafile;
let _dum={ . };
_dum=81|82|99;  @ outlying observations  @
 k1=1;    @ first horizon to generate forecasts @
 k2=24;   @ last horizon  (no. of data points at end not used for estimation) @
 nulobs=seqa(1,1,5); @ observations deleted in AIC computation @
x=0|x;
s=4;
x=trimr(x,rows(x) % s,0);
x=trimr(x,0,k2);
/*
**  Simple adjustment for outliers by averaging over neigbouring observations
**  in corresponding period of the year
*/
_Y=reshape(x,(rows(x)/s),s);            @  pure innovation outliers     @
mu0=meanc(_Y);                          @  mu0  mean uncorrected series @
 if not ismiss(_dum);
  x[_dum]=(x[_dum+s]+x[_dum-s])/2;
 endif;
_Y=reshape(x,(rows(x)/s),s);       @  "simple additive outlier correction" @
T=rows(_Y)*s;
mu1=meanc(_Y);                     @      mu1 mean corrected series        @
mu=mu1;                            @      mu  mean used in estimation      @
_Y=_Y-mu';
_eps=_Y;
/*
**  Simple adjustment for outliers by averaging over neigbouring observations
**  in corresponding period of the year
*/
/*
_Y=reshape(x,(rows(x)/s),s);
mu=meanc(_Y);
*/
/*
 if not ismiss(_dum);
  x[_dum]=(x[_dum+s]+x[_dum-s])/2;
 endif;
_Y=reshape(x,(rows(x)/s),s);
_eps=0;
T=rows(_Y)*s;
mu=meanc(_Y);
_Y=_Y-mu';
*/
/*
**  (Re)Define here esttod for restrictions on the integration parameters;
**      es:     (rx1) the free parameters
**      dest:   (sx1) integration parameters a function of free parameters;
*/
proc esttod(es);
local dest;
@  dest=es*ones(s,1); @
@  dest=es[1]|es[1]|es[2]|es[2];@
   dest=es;
retp(dest);
endp;
/*
** Starting values (here obtained from other estimations)
*/
  dstart=ones(s,1)*0.4;
@  dstart=ones(2,1)*0.4;@
@ dstart=ones(1,1)*0.4;@
dhatt = esttod(dstart);
if rows(dhatt) /= s;
 "Starting values or esttod wrongly specified!";
 waitc;
endif;
/*
**  globals for nroptmum
*/
_nrgtol=1e-9;      /* IMPORTANT: too large value does not work! */
_nroutpu=1;         /* intermediate output shown */
_nralgr=4;          /* POWELL algorithm */
_nrparnm =("d_".*ones(s,1) $+ ftocv(seqa(1,1,s),2,0));
@ _nrparnm="d_12"|"d_34";@ @_nrparnm="d";@
if rows(_nrparnm) /= rows(dstart);
 "_nrparnm wrongly specified!";
 waitc;
endif;
output file = ^outfile reset;
date1=date;
{dhat,sigmatil,gradient,retcode } = nroptmum(&sigdpern,dstart);
comptim=ethsec(date1,date)/100;
output off;
dhatt=esttod(dhat);
/*
** Information criteria
*/
k=s+1+rows(dhat);
if not ismiss(_dum);
 k=k+rows(_dum);  @ no. of parameters @
endif;
nrpar=k;
dimd=rows(dhat);
_eps[nulobs]=zeros(rows(nulobs),1);
noobsaic=rows(_eps)-rows(nulobs);     @ effective number of observations @
sigmaaic=sumc(_eps^2)/noobsaic;
invhess=invpd(_nrfhess);
if scalmiss(invhess);
 "no invertible hessian computed";
endif;
invhess = missrv(invhess[1:dimd,1:dimd],zeros(dimd,dimd));/*invhess not exist*/
covpar = 2*sigmatil*invhess/T; /* approximation of asympt. covariance matrix */
                               /* app. log-likelihood:-(T/2)ln(sigmatil)     */
dvhat = diag(covpar);
/*
plotfunc(&sigdpern,_nrparnm,dhat,dvhat);
*/
/*
**  Forecasting for periodic d
*/
load x[]=^datafile;
 x=0|x;
 n=rows(x);
 if not ismiss(_dum);               /* corrected data used for forecasting */
  x[_dum]=(x[_dum+s]+x[_dum-s])/2;
 endif;
 xc=trimr(x,rows(x) % s,0);
 xc=trimr(xc,0,k2);
 n=n-k2;
 xc=reshape(xc,(rows(xc)/s),s);
 xc2=reshape(x,(rows(x)/s),s);
 n=rows(xc)*s;
 ny=rows(xc);
 muf1=meanc(xc);                @ muf1: means of data used for forecasting @
 @ muf @
 muf=mu;                              @  muf: means used in forecasting    @
 _Y=xc-muf';
 _Yex=xc2-muf';
 /* multistep */
 Yf=fordper(k2,_Y,dhatt);
 nfy=rows(Yf);
 Yf=Yf+muf';
 /* onestep */
 Yres=resdper(_Yex,dhatt);
 Yf2=(_Yex-Yres)+muf';
 yf2=Yf2[ny+1:ny+nfy,.];
 vyf2=vec(Yf2');
 vyf=vec(Yf');
 /*****************    PLOTS ***************************************/
#IFUNIX

let v = 120 120 640 480 0 0 1 6 15 0 0 2 2;
fxy = WinOpenPQG( v, "FORECASTS", "FOREC" );
call WinSetActive( fxy ); WinSetTextWrap(1,1);
#ENDIF

 title("real and forecasts");
 ylabel("rpiinf");
 xlabel("time");
 mat=x[n+k1:n+k2,.]~vyf2~vyf;
 tijd=89+seqa(0,1/s,k2);
 _pltype=6|5|4;
_pcolor=10|12|13;
 _plegctl=1|4|tijd[1]+(maxc(tijd)-tijd[1])*4/5|
          (minc(mat)+0.8*(maxc(mat)-minc(mat)));
 _plegstr="real\0001-step\000multi-step";
 xy(tijd,mat);

/*
#IFUNIX

call WinSetActive( 1 );

#ENDIF
*/
 /***************** RMSE ********************************************/
noforec=k2-k1+1;
xc=reshape(x,(rows(x)/s),s);
rmse1=sqrt((sumc((xc[ny+1:ny+nfy,.]-yf)^2)/noforec*s));
rmse1t=sqrt((sumc(vec((xc[ny+1:ny+nfy,.]-yf)^2))/noforec));
rmse2=sqrt((sumc((xc[ny+1:ny+nfy,.]-yf2)^2)/noforec*s));
rmse2t=sqrt((sumc(vec((xc[ny+1:ny+nfy,.]-yf2)^2))/noforec));
output file = ^outfile on;
"*****************    PARFIMA.PRG       *********************************";
datum=date;tijd=time;
format /rz 5,4;
"Date and time: " (rev(datum[1:3])')~(tijd[1:3]');
"Datafile: " $datafile;
"no. of obs: T:" T;
"no. of effective observations T(AIC):" noobsaic;
if not ismiss(_dum);
 "no. of dummies:" rows(_dum);
 _dum';
else;
"no dummies";
endif;
" number of parameters k:" k;
"starting values dstart:";
dstart';
" convergence tolerance gtol*1e+6:" _nrgtol*1e+6;
nrerror(retcode);
format /rd 12,6;
if maxc(abs(mu-mu0))>1e-8; "Uncorrected data used in estimation d:"; endif;
"seasonal means uncorrected data";
mu0';
"seasonal means corrected data";
mu1';
"seasonal means mu used in estimation";
mu';
"seasonal means of (un)corrected data used in forecasting";
muf1';
"seasonal means mu used in forecasting";
muf';
 PRINTEST(_nrparnm,dhat,dvhat);
"s*1000    s(AIC)*1000     AIC:-T(AIC)*lns(AIC)-2*k";
sigmatil*1000~sigmaaic*1000~(-noobsaic*ln(sigmaaic)-2*k);
"Forecasting" noforec " observations:";
"           RMSE overall and per quarter       ";
"Multiperiod:" rmse1t~(rmse1');
"One-period: " rmse2t~(rmse2');
output off;
"Results in "$+outfile;
